(eval-when (:compile-toplevel :load-toplevel :execute)
;  (require :asdf)
  (require :gsll)
  (require :sb-sprof)
  (require :iterate))
(declaim (optimize (speed 3) (safety 0) (debug 0)))

(defun gref-access (dim)
  "Given an integer dim, this constructs a function that, when supplied with a 
   N-dimensional vector Z and some output vector (-> pointer?), yields the
   corresponding forces"
  (declare (fixnum dim))
  (let ((temp-values (make-array 2 :element-type 'double-float :initial-element 0.0d0)))
    (lambda (zvector output)
	(do ((i 0 (1+ i))) ((= i dim)) (declare (fixnum i))
	  (setf (aref temp-values 0) 0.0d0)
	  (do ((m 0 (1+ m))) ((> m i)) (declare (fixnum m))
	    (do ((n i (1+ n))) ((= n dim)) (declare (fixnum n))
	      (setf (aref temp-values 1) 0.0d0)
	      (do ((k m (1+ k))) ((> k n)) (declare (fixnum k))
		(incf (aref temp-values 1) (grid:gref zvector k)))
	      (incf (aref temp-values 0) (expt (aref temp-values 1) -2))))
	  (setf (grid:gref output i) 
		(- (grid:gref zvector i)
		   (aref temp-values 0)))))))

(defun gref*-access (dim)
  "Given an integer dim, this constructs a function that, when supplied with a 
   N-dimensional vector Z and some output vector (-> pointer?), yields the
   corresponding forces"
  (declare (fixnum dim))
  (let ((temp-values (make-array 2 :element-type 'double-float :initial-element 0.0d0)))
    (lambda (zvector output)
	(do ((i 0 (1+ i))) ((= i dim)) (declare (fixnum i))
	  (setf (aref temp-values 0) 0.0d0)
	  (do ((m 0 (1+ m))) ((> m i)) (declare (fixnum m))
	    (do ((n i (1+ n))) ((= n dim)) (declare (fixnum n))
	      (setf (aref temp-values 1) 0.0d0)
	      (do ((k m (1+ k))) ((> k n)) (declare (fixnum k))
		(incf (aref temp-values 1) (grid:gref* zvector k)))
	      (incf (aref temp-values 0) (expt (aref temp-values 1) -2))))
	  (setf (grid:gref* output i) 
		(- (grid:gref* zvector i)
		   (aref temp-values 0)))))))

(defun cffi-access (dim)
  "Given an integer dim, this constructs a function that, when supplied with a 
   N-dimensional vector Z and some output vector (-> pointer?), yields the
   corresponding forces"
  (declare (fixnum dim))
  (let ((temp-values (make-array 2 :element-type 'double-float :initial-element 0.0d0)))
    (lambda (zvector output)
      (let ((zvector-fptr (grid::foreign-pointer zvector))
	    (output-fptr (grid::foreign-pointer output))
	    )
	(macrolet ((quick-ref (the-vector n)
		     `(cffi:mem-aref
		       ,(case the-vector
			  (zvector 'zvector-fptr)
			  (output 'output-fptr))
		        :double
		       ,n))) 
	  (do ((i 0 (1+ i))) ((= i dim)) (declare (fixnum i))
	    (setf (aref temp-values 0) 0.0d0)
	    (do ((m 0 (1+ m))) ((> m i)) (declare (fixnum m))
	      (do ((n i (1+ n))) ((= n dim)) (declare (fixnum n))
		(setf (aref temp-values 1) 0.0d0)
		(do ((k m (1+ k))) ((> k n)) (declare (fixnum k))
		  (incf (aref temp-values 1) (quick-ref zvector k)))
		(incf (aref temp-values 0) (expt (aref temp-values 1) -2))))
	    (setf (quick-ref output i) 
		  (- (quick-ref zvector i)
		     (aref temp-values 0)))))))))


(let* ((dim 15)
       (vec (grid:make-foreign-array 'double-float :initial-element 1 :dimensions dim))
       (output-1 (grid:make-foreign-array 'double-float :dimensions dim))
       (slow-fn (gref-access dim))
       (intermediate-fn (gref*-access dim))
       (quick-fn (cffi-access dim)))

  (print "gref")
  (time (funcall slow-fn vec output-1))

  (print "gref*")
  (time (funcall intermediate-fn vec output-1))
  (defmethod grid:gref* ((object grid:vector-double-float) linearized-index)
    ;;    (print "modified grid:gref* being used")
    (cffi:mem-aref
     (grid::foreign-pointer object)
     :double
     linearized-index))

  (print "modified gref*")
  (time (funcall intermediate-fn vec output-1))

  (print "hardwired cffi:mem-aref")
  (time (funcall quick-fn vec output-1))
  )

